%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Prepare the workspace
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

home; clear;

% Add path to the BPDQ
addpath(genpath('bpdq_toolbox'))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parameters for saving the results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Save and log results to disk?
saveResults = false;

% If yes create log file and mat file
if(saveResults)
    timestamp = datestr(now, 'yyyy_mmm_dd_HH_MM');
    fileName = sprintf('GenerateUndersamplingPlot_%s.mat', timestamp);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define signal params
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Sparsity
rho = 1/32;

% Signal length
n = 1024;

% Measurement ratio (beta = m/n)
beta = logspace(-0.7, -0.1, 10);

% Number of betas
nBetas = length(beta);

% Number of trials
nTrials = 10;

% Quantization interval
nLevels = 16;

% BPDQ options
p=4;

% For Douglas-Rashford
bpdq_opts={'dr_gamma',.05,'dr_lambda',1.5,'dr_verbosity',0,'fca_verbosity',0};

% Placeholder for all MSEs
mseLMMSE = zeros(nTrials, nBetas);
mseBPDN = zeros(nTrials, nBetas);
mseBPDQ = zeros(nTrials, nBetas);
mseGAMP = zeros(nTrials, nBetas);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute reconstruction errors
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for iBeta = 1:nBetas
    for iTrial = 1:nTrials
        % Number of measurements
        m = round(n*beta(iBeta));
        
        % Generate the signal (Gaussian with proba. rho)
        x = binornd(1, rho, n, 1) .* randn(n, 1) ./ sqrt(rho);
        
        % Generate the measurement matrix
        A = bpdq_generate_sensing_matrix(m, n);
        
        % Obtain measurements1
        z = A*x;
        
        % Quantization interval
        delta = 2.01*norm(z, Inf)/nLevels;
        
        % Quantize measurements
        y = bpdq_quantize(z,delta);
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % LMMSE estimate
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Estimate the variance
        c = var(y - z);
        
        xhatLMMSE = A' * inv((A * A') + (c * eye(m))) * y;
        
        mseLMMSE(iTrial, iBeta) = 10*log10(sum((x-xhatLMMSE).^2)/n);
        
        if(saveResults)
            save(fileName);
        end;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % BPDN - DR
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %epsilon=bpdq_err_p(2,delta,m);
        epsilon = norm(y-z,2);% oracle epsilon
        xhatBPDN  = bpdq_1d(y,A,epsilon,2,bpdq_opts{:});
        
        mseBPDN(iTrial, iBeta) = 10*log10(sum((x-xhatBPDN).^2)/n);
        
        if(saveResults)
            save(fileName);
        end;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % BPDQ - DR
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Let's first compute apropriate epsilon (ref Jacques et. al section III-C)
        %epsilon=bpdq_err_p(p,delta,m);
        epsilon = norm(y-z,p);% oracle epsilon
        xhatBPDQ = bpdq_1d(y,A,epsilon,p,bpdq_opts{:});
        
        mseBPDQ(iTrial, iBeta) = 10*log10(sum((x-xhatBPDQ).^2)/n);
        
        if(saveResults)
            save(fileName);
        end;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % GAMP
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Normalize matrix
        A = A/sqrt(m);
        z = z/sqrt(m);
        
        % Quantization interval
        delta = 2.01*norm(z, Inf)/nLevels;
        
        % Quantize measurements
        y = bpdq_quantize(z,delta);
        
        % Reconstruct
        xhatGAMP = reconstructAmpUndersampled(A, y, 0, rho, delta, x');
        
        mseGAMP(iTrial, iBeta) = 10*log10(sum((x-xhatGAMP').^2)/n);
        
        if(saveResults)
            save(fileName);
        end;
        
        % Notify on command line
        if(mod(iTrial, 1) == 0)
            fprintf('[iB: %d/%d][iT: %d/%d]:\n', iBeta, nBetas, iTrial, nTrials);
            fprintf('\tMSE LMMSE: %.4f dB\n', mseLMMSE(iTrial, iBeta));
            fprintf('\tMSE BPDN-DR: %.4f dB\n', mseBPDN(iTrial, iBeta));
            fprintf('\tMSE BPDQ-DR: %.4f dB\n', mseBPDQ(iTrial, iBeta));
            fprintf('\tMSE GAMP: %.4f dB\n', mseGAMP(iTrial, iBeta));
        end
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Visualize
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% REMOVE NaNs
notNan = ~logical(sum(isnan(mseLMMSE), 2));
mseLMMSE = mseLMMSE(notNan, :);

notNan = ~logical(sum(isnan(mseBPDN), 2));
mseBPDN = mseBPDN(notNan, :);

notNan = ~logical(sum(isnan(mseBPDQ), 2));
mseBPDQ = mseBPDQ(notNan, :);

notNan = ~logical(sum(isnan(mseGAMP), 2));
mseGAMP = mseGAMP(notNan, :);


figure('Color', 'w');
h = semilogx(beta, median(mseLMMSE), '-o', ...
    beta, median(mseBPDN), '--d',...
    beta, median(mseBPDQ), '-.s',...
    beta, median(mseGAMP), 'x--',...
    'LineWidth', 3);
grid on;
set(gca,'FontSize',14);
xlabel('Undersampling (m/n)');
ylabel('MSE (dB)');
legend('Linear MMSE', 'BPDN', 'BPDQ', 'GAMP');
xlim([beta(1), beta(end)]);

% Set colors
co = get(gca,'ColorOrder');
set(h(1),'Color',co(1, :))
set(h(2),'Color',co(2, :))
set(h(3),'Color',co(4, :))
set(h(4),'Color',co(3, :))
